";use strict";
{
  let c;
  let ctx;

  const OPX = 1; // pixels for outline width
  const tracked = new Map();
  const groups = new Map();
  const UPDATE_EVENTS = [
     'load',
     'domcontentloaded',
    'dommousescroll',
    'touchmove',
    'mousewheel',
    'scroll',
    'click',
     'keydown',
    'keyup'
  ];
  const updateEvents = new Set( UPDATE_EVENTS );
  const dct = {
    add_name,
    track, track_all, 
    untrack, untrack_all, 
    start_tracking, pause_tracking,
    guest, restart_tracking
  };

  let tracking = false;

  const tracker = track_elements();

  self.addEventListener('resize', resize_canvas );

  module.exports = dct;

  function add_el( name, el ) {
    const value = groups.get(name);
    if ( !value ) {
      throw new TypeError(`No such name ${name}. Add name and its styles first.` );
    }
    value.els.add(el);
    let tracked_val = tracked.get(el);
    if ( !tracked_val ) {
      tracked_val = { el, names: new Set([name]) };
      tracked.set(el, tracked_val);
    } else {
      tracked_val.names.add(name);
    }
  }

  function add_name( name, style ) {
    let val = groups.get(name);
    if ( !val ) {
      groups.set( name, { styles: new Set([style]), els: new Set() } );
    } else {
      val.styles.add( style );
    }
  }

  function resize_canvas() {
    const style = self.getComputedStyle(c);
    const width = parseFloat(style.width);
    const height = parseFloat(style.height);
    Object.assign( c, {
      width, height
    });
  }

  function guest( isGuest, { sendMessage : sendMessage = null, updateOn : updateOn = UPDATE_EVENTS } = {} ) {
    if ( isGuest ) {
      updateOn.forEach( evName => self.addEventListener( evName, e => update( evName, e ) ) );
    }

    function update( type, e ) {
      const trackedBoxes = getBoxes( tracked );
      const msg = { event : { type, trackedBoxes } };
      if ( !! sendMessage ) {
        sendMessage( msg );
      } else {
        self.top.postMessage( msg, "*" );
      }
    }
  }

  function start_tracking( { absolute : absolute = false, guest : guest = null, updateOn : updateOn = UPDATE_EVENTS } = {} ) {
    if ( !! guest ) {
      self.addEventListener('message', msg => {
        let message = msg.data;
        if ( Array.isArray( message ) ) {
          message = message[0];
        }
        if ( message.event && updateEvents.has( message.event.type ) ) {
          tracker.next( message.event );
        }
      });
    } else {
      updateOn.forEach( evName => self.addEventListener( evName, e => tracker.next(e) ) );
    }
    tracking = true;
    c = document.querySelector('canvas');
    Object.assign( c.style, {
      background: "transparent",
      position: absolute ? "absolute" : "fixed",
      width: "100%",
      height: "100%",
      maxWidth: "100vw",
      maxHeight: "100vh",
      top: 0,
      left: 0,
      pointerEvents: 'none'
    });
    resize_canvas();
    ctx = c.getContext('2d');
    tracker.next({});
  }

  function pause_tracking() {
    tracking = false; 
  }

  function restart_tracking() {
    tracking = true;
    tracker.next();
  }

  function track( selector_or_el, name) {
    let el = selector_or_el;
    if ( typeof selector_or_el == "string" ) {
      el = document.querySelector( selector_or_el );
    }
    add_el(name, el );
  }

  function untrack( sel_or_el, name ) {
    let el = sel_or_el;
    if ( typeof sel_or_el == "string" ) {
      el = document.querySelector( sel_or_el );
    }
    const val = groups.get(name);
    val.els.delete(el);
    const tval = tracked.get(el);
    tval.names.delete(name);
    if ( tval.names.size == 0 ) {
      tracked.delete(el);
    }
  }

  function untrack_all( sel, name ) {
    const els = document.querySelectorAll( sel );
    els.forEach( el => untrack( el, name ) );
  }

  function track_all( selector, name ) {
    const els = document.querySelectorAll( selector );
    els.forEach( el => track( el, name ) );
  }

  function *track_elements() {
    while( true ) {
      if ( tracking ) {
        const e = yield;
        if ( !! e && updateEvents.has( e.type ) ) {
          ctx.clearRect( 0, 0, c.width, c.height );
          let trackedBoxes = e.trackedBoxes;
          if ( ! trackedBoxes ) {
            trackedBoxes = getBoxes( tracked );
          }
          trackedBoxes.forEach( box => paint( box.styles, ...boxToRect( box ) ) );
        }
      } else yield;
    }
  }

  function getBoxes( t ) {
    const trackedBoxes = [...t.values()].reduce( (all_boxes, {el,names}) => {
      const styles = [...names].map( name => groups.get(name).styles ).
        reduce( (all_styles, styles) => ( all_styles.push( ...styles ), all_styles ), [] );
      const rects = Array.from( el.getClientRects() );
      const boxes = rects.map(({width,height,left,top,bottom,right}) => ({styles,width,height,left,top,bottom,right}));
      all_boxes.push( ...boxes );
      return all_boxes;
    }, []);
    return trackedBoxes;
  }

  function paint( styles, x,y,w,h) {
    styles.forEach( s => paint_style( s.style, s.color, x, y, w, h ) );
  }

  function paint_style( style, color, x, y, w, h ) {
    /** use for these ? 
      ctx.save();
      ctx.globalCompositionOperation = 'source-out';
    **/
    if ( style == 'fill' ) {
      ctx.fillStyle = color;
      ctx.fillRect( x, y, w, h );
    } else if ( style == 'outline' ) {
      ctx.lineWidth = 2*OPX;
      ctx.strokeStyle = color;
      ctx.strokeRect( x+OPX,y+OPX,w-2*OPX,h-2*OPX);
    }
  }

  function boxToRect( { left, top, width, height } ) {
    return [ left-1, top-1, width+2, height+2 ];
  }
}
